bits 16
org	0x0


	jmp	BOOT_SEG:start

BOOT_SEG	equ 0x7c0
SETUP_SEG	equ 0x900
KERNEL_SEG	equ 0x920

bootDrv		db 0

sectTrack	dw 18
headCount	dw 2

currentSect	dw 1
readSect	db 0

kernelSect	db KERNEL_SIZE


start:
	mov	ax, cs
	mov	ds, ax
	mov	ss, ax

	mov	sp, 0x1400


	; save boot driver number...
	mov	[bootDrv], dl


	; display system beginning message...
	mov	si, startMsg
	call	put_char


	; load setup file to 0x900:0...
	mov	ax, SETUP_SEG
	mov	es, ax
	mov	ax, [currentSect]
	call	read_sect


	; load kernel image...
	call	load_kernel

	jmp	0x0:0x9000


; load kernel image...
load_kernel:
	mov	ax, KERNEL_SEG
	mov	es, ax

_load_kernel_1:
	mov	ax, [currentSect]	
	call	read_sect
	
	dec	byte[kernelSect]
	jz	_load_kernel
	
	mov	ax, es
	add	ax, 0x20
	mov	es, ax
	
	jmp	_load_kernel_1

_load_kernel:
	ret


; read current sect...
read_sect:
	mov	si, [sectTrack]
	xor	dx, dx
	div	si
	inc	dl
	mov	[readSect], dl
	xor	dx, dx
	div	word[headCount]

	mov	dh, [bootDrv]
	xchg dl, dh
	
	mov	cx, ax
	xchg cl, ch
	shl	cl, 6
	or	cl, [readSect]
	mov	al, 1
	mov	ah, 2
	xor	bx, bx
	int	0x13
	jc	boot_fail
	inc	byte[currentSect]
	ret


; display message...
put_char:
	lodsb
	or	al, al
	jz	_put_char

	mov	ah, 0xe
	mov	bx, 0x7
	int	0x10
	jmp	put_char

_put_char:
	ret


; boot failed...
boot_fail:
	mov	si, failMsg
	call	put_char
_fail:
	jmp	_fail


startMsg	db 'Start MyOs...', 0xa, 0xd, 0x0
failMsg		db 'Boot failed...', 0x0


	; mbr flag set...
	times 510-($-$$) db 0
	dw 0xaa55
